﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Timers;
using System.Threading;
using System.Threading.Tasks;
using System.IO;
using System.IO.Pipes;

namespace MetadataSDK
{
    

    /// <summary>
    /// Handles creating, monitoring, and using the Named Pipe server stream.
    /// </summary>
    /// <remarks>
    /// There are three threads at work here.<br />
    /// <br />
    /// 1. The main UI thread (U) which creates the PipeMessenger instance and calls Start().
    ///    Start() uses QueueUserWorkItem to wait for a connection in the background.<br /> 
    /// 2. Once a connection is made, the messenger thread (M) is spawned which is responsible for
    ///    reading and writing data to the pipe.<br />
    /// 3. The heartbeat timer (H) is also started which periodically checks that there is still a client
    ///    connected to the pipe.<br /> 
    /// <br />  
    ///    (H) is required because:<br />
    ///      A. (M) blocks forever on read().<br /> 
    ///      B. There is no built-in mechanism in named pipes to notify us when the client has disconnected.<br />
    ///      C. A client can't reconnect to the pipe until the server allows it.<br />
    ///<br />
    ///    These three facts create the situation where if we're reading from a pipe, but the client disconnects, 
    ///    our app will just wait forever to hear from the client, while the client keeps failing to reconnect
    ///    to our pipe.<br />
    ///<br />
    ///    So (H) checks if the client is still connected. If it isn't, it kills (M), and reestablishes the pipe,
    ///    allowing the client to reconnect. 
    /// </remarks>
    class PipeMessenger
    {
        volatile bool _keepRunning;
        volatile bool _isConnected;

        Thread _messengerThread;
        System.Timers.Timer _heartbeat;

        string _pipename;
        object _pipelock = new object();
        NamedPipeServerStream _pipe;

        const int kHeartbeatInterval = 4000;
        const char kMessageDelimeter = ';';
        const int kReadBufferMax = 32768;

        Func<string, string> _messageCallback;
        
        /// <summary>
        /// Creates a new instance of the PipeMessenger class
        /// </summary>
        /// <param name="pipename">The name of the pipe to open 
        /// (should be in the form {process id}_airfoil_metadata)</param>
        /// <param name="callback">The function we should call when a message is 
        /// received over the pipe. We use a functional model here rather 
        /// than event-based, because the protocol is synchronous.</param>
        public PipeMessenger(string pipename, Func<string, string> callback)
        {
            _pipename = pipename;
            _messageCallback = callback;
        }

        /// <summary>
        /// Starts the pipe messenger. This attempts to open the pipe via a background worker thread,
        /// and starts the heartbeat timer (H).
        /// </summary>
        public void Start()
        {
            _keepRunning = true;
            TryConnect();

            // The heartbeat timer just checks if the pipe is connected, since there's no built-in
            // mechanism for a named pipe to notify us if the client disconnects. 
            // If the client is disconnected, we kill the messenger thread and restart the connection.
            _heartbeat = new System.Timers.Timer();
            _heartbeat.Elapsed += HeartDidBeat;
            _heartbeat.Interval = kHeartbeatInterval;
            _heartbeat.Start();
        }

        /// <summary>
        /// Close the pipe and stop the heartbeat timer (H).
        /// </summary>
        public void Stop()
        {
            _keepRunning = false;
            ClosePipe();

            _heartbeat.Stop();
            _heartbeat.Dispose();
        }

        /// <summary>
        /// Disconnect and dispose of the pipe and stop the messenger thread (M). 
        /// Ignores any exceptions.
        /// </summary>
        void ClosePipe()
        {
            lock (_pipelock)
            {
                try
                {
                    if (_pipe != null)
                    {
                        _pipe.Disconnect();
                        _pipe.Dispose();
                    }
                }
                catch (System.Exception) { }
            }

            // Kill the messenger
            if (_messengerThread != null)
            {
                _messengerThread.Abort();
                _messengerThread = null;
            }
        }

        /// <summary>
        /// Performs the actual pipe creation and waits for a client connection. 
        /// This method will block until a connection is made, so it's best to run it
        /// in the background.
        /// </summary>
        void Connect()
        {
            // Open the pipe and wait for a connection. Will block forever, so call in the background.
            _pipe = new NamedPipeServerStream(_pipename);
            _pipe.WaitForConnection();
            _isConnected = true;
        }

        /// <summary>
        /// Runs Connect() in the background. Once connection is established, 
        /// the background worker thread also spawns the messenger thread (M)
        /// which reads messages from the pipe.
        /// </summary>
        void TryConnect()
        {
            System.Threading.ThreadPool.QueueUserWorkItem(delegate
            {
                lock (_pipelock)
                {
                    ClosePipe();
                    Connect();            
                }

                // Connected!
                if (_isConnected)
                {
                    _messengerThread = new Thread(MessengerLoop);
                    _messengerThread.IsBackground = true;
                    _messengerThread.Start();
                }

            }, null);
        }

        /// <summary>
        /// Called by the heartbeat timer every kHeartbeatInterval milliseconds. 
        /// </summary>
        /// <param name="sender">The timer instance.</param>
        /// <param name="e">Timer elapsed args.</param>
        void HeartDidBeat(object sender, ElapsedEventArgs e)
        {
            CheckConnection();
        }

        /// <summary>
        /// If we think we're still connection, this method checks that the 
        /// client is actually connected to the pipe. 
        /// </summary>
        /// <remarks>
        /// Note the use of Monitor.TryEnter
        /// here, which attempts to lock the pipe to perform this check. Pipes are not threadsafe,
        /// but we don't want to wait if we the pipe is busy. We'll just try again at the next
        /// heartbeat interval.<br />
        /// <br />
        /// If the client isn't connected, we close the pipe and retry the connection in the
        /// background using TryConnect().
        /// </remarks>
        void CheckConnection()
        {
            if (!_isConnected)
                return;

            bool lockTaken = false;
            try
            {
                Monitor.TryEnter(_pipelock, ref lockTaken);
                if (lockTaken && !_pipe.IsConnected)
                {
                    _isConnected = false;
                    ClosePipe();

                    // Retry the connection
                    TryConnect();
                }
            }
            finally
            {
                if (lockTaken)
                    Monitor.Exit(_pipelock);
            }
        }

        /// <summary>
        /// This is the method run by the messenger thread (M). It continually attempts to read
        /// data from the pipe and parse the reads into messages. 
        /// </summary>
        /// <remarks>
        /// Since messages are sent in the format length;message with no delimiter in-between 
        /// messages, we first need to read until we get the length, then use that to read in the
        /// full message.<br />
        /// <br />
        /// There are alternative models for framing these messages of course:<br />
        /// <br />
        /// 1. Using a delimter between messages. We opted not to do this, because some 
        /// implementations would need to know the message size anyway in order to properly allocate
        /// a buffer. Having a message length and a message delimiter overcomplicated things.<br />
        /// <br />
        /// 2. Putting the pipe into messaging mode. This guarantees that every write to the pipe is
        /// a discrete message and while reading we can ask the pipe to read until the message is fully
        /// retrieved. Unfortunately, not all implementations of named pipes support this mode.<br />
        /// <br />
        /// Once messages are received, we call the message callback function, providing the
        /// message and receiving a response. That response is then sent over the pipe.<br />
        /// <br />
        /// Note that we do not use locking for reads and writes. This is because if we were to
        /// lock on read it would block forever and the heartbeat timer would never be able to get a lock
        /// and check the pipe connection.
        /// </remarks>
        void MessengerLoop()
        {
            try
            {
                StringBuilder messageBuffer = new StringBuilder();
                byte[] readBuffer = new byte[kReadBufferMax];
                int remainingMessageLength = 0;

                while (_isConnected && _keepRunning)
                {
                    // Get Next message Size
                    if (remainingMessageLength == 0)
                    {
                        char theByte;
                        theByte = (char)_pipe.ReadByte();

                        if (theByte == kMessageDelimeter)
                        {
                            string messageLength = messageBuffer.ToString();
                            remainingMessageLength = Int32.Parse(messageLength);
                            messageBuffer.Clear();
                        }
                        else
                        {
                            messageBuffer.Append(theByte);
                        }
                    }
                    else
                    {
                        // If we know the size, pull in as much of the message as we can
                        int shouldReadLength = Math.Min(kReadBufferMax, remainingMessageLength);
                        int readLength;

                        readLength = _pipe.Read(readBuffer, 0, shouldReadLength);

                        remainingMessageLength -= readLength;

                        string messageFragment = Encoding.UTF8.GetString(readBuffer, 0, readLength);
                        messageBuffer.Append(messageFragment);

                        if (remainingMessageLength == 0)
                        {
                            string fullMessage = messageBuffer.ToString();
                            messageBuffer.Clear();

                            // We have a message, but nobody cares
                            if (_messageCallback == null)
                                continue;

                            string response = _messageCallback(fullMessage);

                            // No response sent
                            if (response == null)
                                continue;

                            // Send the response
                            int messageLength = Encoding.UTF8.GetByteCount(response);
                            string framedMessage = String.Format("{0};{1}", messageLength, response);
                            byte[] messageBytes = Encoding.UTF8.GetBytes(framedMessage);

                            _pipe.Write(messageBytes, 0, messageBytes.Length);
                            _pipe.Flush();
                        }
                    }
                }
            }
            catch(System.Threading.ThreadAbortException)
            {
                
            }
            catch(System.IO.IOException)
            {

            }
        }

    }
}
